#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
"""Version number for the mlos_viz package."""

# NOTE: This should be managed by bumpversion.
VERSION = "0.6.1"

if __name__ == "__main__":
    print(VERSION)
